
#include <NetAddress.h>
#include <NetEndpoint.h>
#include <StopWatch.h>
#include <File.h>
#include <stdio.h>
#include <stdlib.h>
#include <socket.h>
#include <OS.h>
#include <string.h>

//#define DEBUG
//#define BARGRAPH

#define THREADED

#define MAX_PPT		(256)
#define MIN_PORT	(1)
#define MAX_PORT	(1024)

/* Version 3 Portscan */
/* basically we will create a large structure that each thread has to access to find out who to scan and what port */
/* each thread  will lock the structure only long enough to get source, port then release, do the scan then reacquire lock for results. */

/* global structure to lock, and lock itself */
char ports[MAX_PORT];
sem_id gLock;

/* data that will be passed to each thread */
typedef struct _tStruct {
	char dest[40];
	short int baseport, maxport;
	} tStruct;

tStruct ScanData[(MAX_PORT/MAX_PPT)];

/* global thread data */
thread_id threads[(MAX_PORT/MAX_PPT)];

/* prototype of thread function */
int32 PortScan(void *data);

int main(int argc, char *argv[])
{
	
	int i,j;
	
	#ifdef THREADED
		char string[B_OS_NAME_LENGTH];
		status_t rc;
	#endif
	#ifdef BARGRAPH
		char bargraph[(MAX_PORT/MAX_PPT)];
		char UpdateLine[30];
	#endif

	printf("portscan by Erik Rainey\n");

	#ifdef DEBUG
	printf("Sizeof(tStruct)=%li\n",sizeof(tStruct));
	#endif
		
	if(argc < 2)
	{
		printf("Insuffient number of arguments\n");
		printf("usage: $ portscan <hostname or IP> [<hostname or IP> ...]\n");
		printf("Up to and including 9 destinations may be specified.\n");
		exit(1);
	}
	
	if(MAX_PORT%MAX_PPT!=0)
		printf("LATE COMPILE ERROR: Number of ports to scan is not evenly divisible by number of ports per thread to scan\n");

	/* start the clock! */
	BStopWatch *watch = new BStopWatch("Timer 0");

	gLock = create_sem(1,"Global Ports Array Lock");

	printf("A report will be generated when each destination scan is finished.\n");
	for(i=1;i<10;i++)
	{
    	// don't process a NULL string
		if(argv[i]==NULL)
			break;
			
		/* clear ports structure */
		memset(ports,0,MAX_PORT);

		/* number of threads to spawn */
		for(j=0;j<(MAX_PORT/MAX_PPT);j++)
		{
			/* this is the destination */
			strcpy(ScanData[j].dest,argv[i]);
			/* this is the range of port numbers this thread will scan */
			ScanData[j].baseport = j*MAX_PPT;
			ScanData[j].maxport  = j*MAX_PPT + MAX_PPT - 1 ;
			#ifdef THREADED
				/* name the thread */
				sprintf(string,"portscan %s %i:%i",argv[i],ScanData[j].baseport,ScanData[j].maxport);
				/* spawn it */
				threads[j] = spawn_thread(PortScan,string,B_NORMAL_PRIORITY,(void *)j);
				#ifdef DEBUG
				printf("Created %li:%s\n",threads[j],string);
				#endif
				/* resume the thread */
				resume_thread(threads[j]);
			#else
				PortScan((void *)j);
			#endif
		}
		
		#ifdef THREADED
		#ifdef BARGRAPH
			/*prepare bargraphs */
			memset(bargraph,' ',10);
			bargraph[10] = '\0';

			/* print out initial graph */
			sprintf(UpdateLine, "Threads Scanning [%s]\r",bargraph);
			printf("%s",UpdateLine);
			fflush(stdout);
		
			/* now wait for all the threads to die */
			for(j=0;j<(MAX_PORT/MAX_PPT);j++)
			{	
				wait_for_thread(threads[j],&rc);
				if(rc==B_OK)
					bargraph[j] = '.';
				else
					bargraph[j] = 'x';
				sprintf(UpdateLine, "Threads Scanning [%s]\r",bargraph);
				printf("%s",UpdateLine);
				fflush(stdout);
			}
			printf("\n");
		#else
			for(j=0;j<(MAX_PORT/MAX_PPT);j++)
				wait_for_thread(threads[j],&rc);		
		#endif
		#endif
		
		
		/* now print out info */
		printf("%s: ",argv[i]);
		for(j=0;j<MAX_PORT;j++)
			if(ports[j]==1)
				printf("%i ",j);
		printf("\n");
	}
		
	delete_sem(gLock);	
		
	delete watch;
		
	return 0;
}	


int32 PortScan(void *data)
{
	BNetEndpoint bnep;
	int dport;
	thread_id this_thread;
	tStruct ldata;
	int lports[MAX_PPT];
	int i,j; 	// j is the index into the shared memory pool
	
	j = (int)data;
	this_thread = find_thread(NULL);
	memcpy(&ldata,&ScanData[j],sizeof(tStruct));
	
	#ifdef DEBUG
		printf("Parameter=%i\n",j);
		printf("ScanData (dest:%s bp:%i mp:%i)\n",ScanData[j].dest,ScanData[j].baseport,ScanData[j].maxport);
		printf("ldata    (dest:%s bp:%i mp:%i)\n",ldata.dest,ldata.baseport,ldata.maxport);
	#endif
	
	if(bnep.InitCheck() < B_OK)
	{
		printf("ERROR!!!\n");
		return B_ERROR;
	}
	
	/* can't scan port 0 */
	if(ldata.baseport==0)
	{
		dport = 1;
		i = 1;
		lports[0] = 0;
	}
	else
	{
		dport = ldata.baseport;
		i = 0;
	}
		
	for(/*see above*/;dport<=ldata.maxport;dport++,i++)
	{
		#ifdef DEBUG
		printf("%li:@%i\r",this_thread,dport);
		fflush(stdout);
		#endif
		// get host info
		if(bnep.Connect(ldata.dest,dport)==B_OK)
		{
			lports[i] = 1;	
			bnep.Close();
		}
		else
			lports[i] = 0;
	}
	//printf("%li: Finished Scanning\n",this_thread);
	
	#ifdef THREADED
		/* lock global structure */
		acquire_sem(gLock);
	#endif
	
	#ifdef DEBUG
	printf("%li: Updating ports[]\n",this_thread);
	#endif
		
	/* update array */
	for(i=0,j=ldata.baseport;j<=ldata.maxport;i++,j++)
		ports[j] = lports[i];
			
	#ifdef THREADED
		/* unlock global structure */
		release_sem(gLock);	
	#endif
	
	return B_OK;
}

